// Description: An open-source NVC (and NEC V810) disassembler.

// History:
// 8 Apr 1998 - Created
// 9 Apr 1998 - Version 1.0 completed
// Many, many changes since then!
// v1.04 24 Nov 1998
// Quite a few more changes...
// v1.05 Mar 2010 - See readme for change log (AJB)
// v???? Aug-Oct 2015 - See readme for change log (MH)

// Authors:
// Bob VanderClay (vandercl@umich.edu)
// David Tucker (dbt@dana.ucc.nau.edu)
// Amos Bieler (abieler@gmx.net)
// Matej Horvat (matej.horvat@guest.arnes.si)

#include <ctype.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "mnvcdis.h"

// Function name-address pair structure
typedef struct stFuncName
{
  struct stFuncName* Next;
  word Address;
  unsigned int ReferenceCount;
  char Name[];
} tyFuncName;

word PC;
word PCEnd;
FILE* InFile;
FILE* OutFile;
FILE* MapFile;
FILE* CommentFile;
tyFuncName* FuncNameList;

/* Routines to automate the mundane task of sign extending... */
word sign_26 (word num)
{
  if(num & 0x02000000)
    return ((word)num | 0xFC000000);
  return(num);
}

word sign_16 (hword num)
{
  if(num & 0x8000)
    return ((word)num | 0xFFFF0000);
  return(num);
}

word sign_9 (hword num)
{
  if(num & 0x0100)
    return ((word)num | 0xFFFFFE00);
  return(num);
}

word sign_8 (byte num)
{
  if(num & 0x0080)
    return ((word)num | 0xFFFFFF00);
  return(num);
}

word sign_5 (byte num)
{
  if(num & 0x0010)
    return ((word)num | 0xFFFFFFE0);
  return(num);
}

char* GPRName (int Which)
{
  // Takes a GPR register number and returns its name or number in string
  // form.

  // We need two buffers because this function may be called twice for a
  // single printf call.
  static char BufferA[3];
  static char BufferB[3];
  static bool WhichBuffer = false;

  if(Which == 3)
    return "SP";
  if(Which == 4)
    return "GP";
  if(Which == 31)
    return "LP";

  WhichBuffer = !WhichBuffer;

  sprintf(WhichBuffer ? BufferA : BufferB, "%d", Which);

  return WhichBuffer ? BufferA : BufferB;
}

char* SRName (int Which)
{
  // Takes a system register number and returns its name.

  static char Buffer[3];

  switch(Which)
  {
    case 0: return "EIPC";
    case 1: return "EIPSW";
    case 2: return "FEPC";
    case 3: return "FEPSW";
    case 4: return "ECR";
    case 5: return "PSW";
    case 6: return "PIR";
    case 7: return "TKCW";
    case 24: return "CHCW";
    case 25: return "ADTRE";
  }

  sprintf(Buffer, "%d", Which);

  return Buffer;
}

char* FuncName (word Address)
{
  // Takes an address and returns a function name or the address in
  // hexadecimal string form.

  tyFuncName* Iterator;

  static char Buffer[9];

  Iterator = FuncNameList;
  while(Iterator)
  {
    if(Iterator->Address == Address)
    {
      Iterator->ReferenceCount++;
      return Iterator->Name;
    }
    Iterator = Iterator->Next;
  }

  sprintf(Buffer, "%X", Address);

  return Buffer;
}

void Quit (unsigned int ErrorCode)
{
  // Frees allocated memory, closes open files, and quits.

  tyFuncName* FuncName;
  tyFuncName* NextFuncName;

  FuncName = FuncNameList;
  while(FuncName)
  {
    NextFuncName = FuncName->Next;
    free(FuncName);
    FuncName = NextFuncName;
  }

  if(InFile)
    fclose(InFile);
  if(OutFile)
    fclose(OutFile);
  if(MapFile)
    fclose(MapFile);

  exit(ErrorCode);
}

void Crash (char* Message)
{
  // Prints a message to the standard error stream and quits.

  if(Message)
    fputs(Message, stderr);

  Quit(1);
}

void ReadFuncNames (char* MapFileName)
{
  // Takes a map file name and reads function names from it.
  // This code is ugly and should be improved.

  tyFuncName* FuncName;
  char Buffer[32];
  word Address;
  unsigned int Length;
  int Ch;

  MapFile = fopen(MapFileName, "r");
  if(!MapFile)
    Crash("Cannot open function map file");

  for(;;)
  {
    Ch = fgetc(MapFile);
    if(Ch == EOF)
      break;

    // If first character in line is a semicolon, it's a comment
    if(Ch == ';')
    {
      for(;;)
      {
        Ch = fgetc(MapFile);
        if(Ch == '\n' || Ch == EOF)
          break;
      }
      if(Ch == '\n')
        continue;
      if(Ch == EOF)
        break;
    }
    else if(Ch == '\n')
      // Blank line
      continue;

    // Get address
    Length = 1;
    Buffer[0] = (char)Ch;
    for(;;)
    {
      Ch = fgetc(MapFile);
      if(Ch == EOF || isblank(Ch))
        break;
      if(Length == 8)
        Crash("Too long address in function map file");
      Buffer[Length++] = (char)Ch;
    }
    if(Ch == EOF)
    {
      if(Length)
        fputs("Expected whitespace in function map file", stderr);
      break;
    }
    Buffer[Length] = '\0';
    sscanf(Buffer, "%X", &Address);

    // Skip whitespace
    while(isblank(Ch))
      Ch = fgetc(MapFile);
    if(Ch == EOF)
    {
      fputs("Expected a function name in function map file", stderr);
      break;
    }

    // Get function name
    Length = 1;
    Buffer[0] = (char)Ch;
    for(;;)
    {
      Ch = fgetc(MapFile);
      if(Ch == '\n' || Ch == EOF)
        break;
      if(Length == sizeof(Buffer) - 1)
        Crash("Too long function name in function map file");
      Buffer[Length++] = (char)Ch;
    }
    Buffer[Length] = '\0';

    // Allocate new structure
    FuncName = (tyFuncName*)malloc(sizeof(tyFuncName) + Length + 1 /* NUL */);
    if(!FuncName)
      Crash("Out of memory while reading function map file");
    FuncName->Address = Address;
    FuncName->ReferenceCount = 0;
    strcpy(FuncName->Name, Buffer);

    // Add to beginning of linked list (it's easier that way)
    FuncName->Next = FuncNameList;
    FuncNameList = FuncName;
  }
}

void PrintFuncStats ()
{
  // Prints function statistics to standard output.

  tyFuncName* Iterator;

  Iterator = FuncNameList;
  while(Iterator)
  {
    if(!(Iterator->ReferenceCount))
      printf("Function %s is never called\n", Iterator->Name);
    Iterator = Iterator->Next;
  }
}

void Disassemble (unsigned int InstructionCount)
{
  // Disassembles the given number of instructions.

  char lowB, highB, lowB2, highB2;  /* up to 4 bytes for instruction (16 or 32 bits) */
  int opcode, arg1, arg2, arg3;

  while(PC < PCEnd && InstructionCount--)
  {
    fseek(InFile, PC, SEEK_SET);
    lowB = fgetc(InFile);
    highB = fgetc(InFile);
    lowB2 = fgetc(InFile);
    highB2 = fgetc(InFile);

    if(feof(InFile))
      break;

    /* Aren't these trapped somewhere below? -ajb */
    if(
      ((highB == 0xFF) && (lowB == 0xFF)) ||
      ((highB == 0x00) && (lowB == 0x00))
    )
    {
      fprintf(OutFile, "%08X\t?db 0x%02X, 0x%02X\n", PC, lowB, highB);
      PC += 2;
      continue;
    }

    opcode = highB >> 2;

    /* Special opcode format for type III (branch) instructions. */
    if((highB & 0xE0) == 0x80)
      opcode = (highB >> 1);

    /* Error Invalid opcode! */
    if(opcode > 0x4F | opcode < 0)
    {
      fprintf(OutFile, "%08X\t?db 0x%02X\n", PC, lowB);
      PC++;
      continue;
    }

    switch (optable[opcode].addr_mode)
    {
      case AM_I:
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        if(opcode == JMP)
          fprintf(OutFile, "%08X\t%s $%s\n", PC, optable[opcode].opname, GPRName(arg2));
        else
          fprintf(OutFile, "%08X\t%s $%s, $%d\n", PC, optable[opcode].opname, GPRName(arg2), arg1);
        PC += 2;  /* 16 bit instruction */
        break;

      case AM_IIa:
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        if(opcode == LDSR)
          fprintf(OutFile, "%08X\t%s $%s, $%s\n", PC, optable[opcode].opname, GPRName(arg1), SRName(arg2));
        else if(opcode == STSR)
          fprintf(OutFile, "%08X\t%s $%s, $%s\n", PC, optable[opcode].opname, SRName(arg2), GPRName(arg1));
        else
          fprintf(OutFile, "%08X\t%s %d, $%s\n", PC, optable[opcode].opname, sign_5(arg2), GPRName(arg1));
        PC += 2;  /* 16 bit instruction */
        break;

      case AM_IIb:
        // TODO SETF mnemonics
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        fprintf(OutFile, "%08X\t%s %d, $%s\n", PC, optable[opcode].opname, arg2, GPRName(arg1));
        PC += 2;  /* 16 bit instruction */
        break;

      case AM_IIc:  /* Bit-string op; get subcode */
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        if(arg2 > 16)
          fprintf(OutFile, "%08X\t; Invalid bitstring subop 0x%02X\n", PC, arg2);
        else
          fprintf(OutFile, "%08X\t%s\n", PC, bssuboptable[arg2]);
        PC += 2;  /* 16 bit instruction */
        break;

      case AM_IId:
        fprintf(OutFile, "%08X\t%s\n", PC, optable[opcode].opname);
        PC += 2;  /* 16 bit instruction */
        break;

      case AM_IIe:
        arg1 = (lowB & 0x1F);
        fprintf(OutFile, "%08X\t%s 0x%02X\n", PC, optable[opcode].opname, arg1);
        PC += 2;  /* 16 bit instruction */
        break;

      case AM_III:
        arg1 = ((highB & 0x1) << 8) + (lowB & 0xFE);
        fprintf(OutFile, "%08X\t%s %X\n", PC, optable[opcode].opname, PC + sign_9(arg1));
        PC += 2;  /* 16 bit instruction */
        break;

      case AM_IV:
        arg1 = ((highB & 0x3) << 24) + (lowB << 16) + (highB2 << 8) + lowB2;
        fprintf(OutFile, "%08X\t%s %s\n", PC, optable[opcode].opname, FuncName(PC + sign_26(arg1)));
        PC += 4;  /* 32 bit instruction */
        break;

      case AM_V:
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        arg3 = (highB2 << 8) + lowB2;
        fprintf(OutFile, "%08X\t%s 0x%X, $%s, $%s\n", PC, optable[opcode].opname, arg3, GPRName(arg2), GPRName(arg1));
        PC += 4;  /* 32 bit instruction */
        break;

      case AM_Vb:
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        arg3 = (highB2 << 8) + lowB2;
        fprintf(OutFile, "%08X\t%s 0x%X, $%s, $%s\n", PC, optable[opcode].opname, sign_16(arg3), GPRName(arg2), GPRName(arg1));
        PC += 4;  /* 32 bit instruction */
        break;

      case AM_VIa:  /* Mode6 form1 */
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        arg3 = (highB2 << 8) + lowB2;
        fprintf(OutFile, "%08X\t%s %d[$%s], $%s\n", PC, optable[opcode].opname, sign_16(arg3), GPRName(arg2), GPRName(arg1));
        PC += 4;  /* 32 bit instruction */
        break;

      case AM_VIb:  /* Mode6 form2 */
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        arg3 = (highB2 << 8) + lowB2;
        fprintf(OutFile, "%08X\t%s $%s, %d[$%s]\n", PC, optable[opcode].opname, GPRName(arg1), sign_16(arg3), GPRName(arg2));
        PC += 4;  /* 32 bit instruction */
        break;

      case AM_VII:  /* Extended instruction; get subcode */
        arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
        arg2 = (lowB & 0x1F);
        arg3 = (highB2 >> 2);
        if(arg3 > 16)
          fprintf(OutFile, "%08X\t; Invalid FPU subop 0x%02X\n", PC, arg3);
        else
          fprintf(OutFile, "%08X\t%s\t$%s,\t$%s\n", PC, fpsuboptable[arg3], GPRName(arg1), GPRName(arg2));
        PC += 4;  /* 32 bit instruction */
        break;

      /* Everything else is probably data. */
      case AM_RFU:
      default:
        fprintf(OutFile, "%08X\t?db 0x%02X\n", PC, lowB);
        PC++;
    }
  }
}

int main (int argc, char* argv[])
{
  word Length;
  bool CfgPrintFuncStats;
  unsigned int i;
  unsigned int InstructionCount;
  int Ch;

  /* Check parameter count. */
  if(argc < 3)
  {
    puts(
      "V810/Virtual Boy disassembler\n"
      "\n"
      "Usage: MNVCDIS inFile outFile [start [{+length | end} [options]]]\n"
      "\n"
      "Options (case insensitive):\n"
      "\n"
      "  C commentFile  The file containing comments to merge with the\n"
      "                 disassembly.\n"
      "  FM mapFile     The file containing function address-name pairs (one\n"
      "                 per line, separated by whitespace).\n"
      "  FS             Prints function statistics (requires FM).\n"
      "\n"
      "If outFile is \"-\", standard output will be used.\n"
      "All numbers are in hexadecimal."
    );
    return 0;
  }

  PC = 0;
  PCEnd = (word)0xFFFFFFFF;
  InFile = NULL;
  OutFile = NULL;
  MapFile = NULL;
  CommentFile = NULL;
  FuncNameList = NULL;
  CfgPrintFuncStats = false;

  if(argc > 3)
    sscanf(argv[3], "%X", &PC);

  if(argc > 4)
    if(argv[4][0] == '+')
    {
      sscanf(&(argv[4][1]), "%X", &Length);
      PCEnd = PC + Length;
    }
  else
    sscanf(argv[4], "%X", &PCEnd);

  if(argc > 6)
    for(i = 5; i < argc; i++)
      if(toupper(argv[i][0]) == 'C')
      {
        CommentFile = fopen(argv[++i], "r");
        if(!CommentFile)
          Crash("Cannot open comment file");
      }
      else if(toupper(argv[i][0]) == 'F' && toupper(argv[i][1]) == 'M')
        ReadFuncNames(argv[++i]);
      else if(toupper(argv[i][0]) == 'F' && toupper(argv[i][1]) == 'S')
        CfgPrintFuncStats = true;

  /* Open input file. */
  InFile = fopen(argv[1], "rb");
  if(!InFile)
    Crash("Cannot open input file");

  /* Open output file (or use standard output). */
  if(argv[2][0] == '-' && argv[2][1] == '\0')
    OutFile = stdout;
  else
    OutFile = fopen(argv[2], "w");

  if(!OutFile)
    Crash("Cannot open output file");

  if(CommentFile)
    for(;;)
    {
      Ch = fgetc(CommentFile);
      if(Ch == EOF)
        break;
      else if(Ch == ';' || Ch == '\n')
      {
        fputc(Ch, OutFile);
        if(Ch == '\n')
          continue;

        for(;;)
        {
          Ch = fgetc(CommentFile);
          if(Ch == '\n' || Ch == EOF)
            break;
          fputc(Ch, OutFile);
        }
        fputc('\n', OutFile);
      }
      else
      {
        InstructionCount = 0;
        for(;;)
        {
          InstructionCount =
            InstructionCount * 10 + (unsigned int)((char)Ch - '0');
          Ch = fgetc(CommentFile);
          if(Ch == '\n' || Ch == EOF)
            break;
        }

        Disassemble(InstructionCount);
      }
    }
  else
    Disassemble((unsigned int)0xFFFFFFFF);

  if(CfgPrintFuncStats)
    PrintFuncStats();

  Quit(0);

  return 0;
}
